/*
 * MixedMatrixP.cpp	(VIAP)
 * Copyright (C) 2004 Yuji MATSUOKA
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *
 *
 * Yuji MATSUOKA:
 *   mailto:yuji_matsuoka@mist.i.u-tokyo.ac.jp
 *   http://www.sr3.t.u-tokyo.ac.jp/~matsuoka/
 */

#include "MixedMatrixP.hpp"

#define INFINITY 100000000

MixedMatrixP::MixedMatrixP()
{
	row = 0;
	col = 0;
	flag = true;
	rank = 0;
	weight = 0;
	delta = NULL;
	rDims = NULL;
	cDims = NULL;
	
	phi = NULL;
	psi = NULL;
	
	vnum = 0;
	BTCnum = 0;
	BCnum = 0;
	BTnum = 0;
	BPnum = 0;
	BPmax = 0;
	BP1num = 0;
	BP2num = 0;
	BP3num = 0;
	
	BTChead = NULL;
	BTCadjlist = NULL;
	BTCvfirst = NULL;
	BTCweight = NULL;
	BTCpweight = NULL;
	BPhead = NULL;
	BPadjlist = NULL;
	BPvfirst = NULL;
	BPweight = NULL;
	vattr = NULL;
	
	p = NULL;
	d = NULL;
	path = NULL;
	pathLinfo = NULL;
	prev = NULL;
	previnfo = NULL;
}
void MixedMatrixP::init(int r, int c)	// r : number of rows, c : number of columns
{
	row = r;
	col = c;
	rank = 0;
	freeMemory();
	allocMemory();
	matQ.init(row, col);
	matT.init(row, col);
}
void MixedMatrixP::readFromStream(istream &in)
{
	int r, c;
	in >> r >> c;					// r : number of rows, c : number of columns
	init(r,c);
	
	cout << "row = " << row << "\n";
	cout << "col = " << col << "\n";
	cout << "\n";
	
	for(int i = 0; i < row; i++){				// input of Q
		for(int j = 0; j < col; j++){
			int n;
			in >> n;
			RealNumber rn(n);
			matQ.put(i, j, rn);					// coefficient of element
			int d;
			in >> d;
			matQ.putDim(i, j, d);					// dimension of element
		}
	}
	for(int i = 0; i < row; i++){				// input of T
		for(int j = 0; j < col; j++){
			int n;
			in >> n;
			Variable var(n);					// n > 0 means nonzero element
			matT.put(i, j, var);
			int d;
			in >> d;
			matT.putDim(i, j, d);					// dimension of element
		}
	}
	
	if(!checkDim()){ // check the assumption on matrix Q
		cout << "incorrect input !";
		exit(0);
	}
}
void MixedMatrixP::writeToStream(ostream &out)
{
	for(int i = 0; i < row; i++){
		for(int j = 0; j < col; j++){
			out << setw(5) << matQ.get(i,j);
			if(matQ.get(i,j).isZero()){
				out << "    ";
			} else {
				out << "s^" << matQ.getDim(i,j) << " ";
			}
		}
		out << "\n";
	}
	out << "\n";
	for(int i = 0; i < row; i++){
		for(int j = 0; j < col; j++){
			out << setw(5) << matT.get(i,j);
		}
		out << "\n";
	}
	out << "\n";
	for(int i = 1; i <= row; i++){
		out << rDims[i] << " ";
	}
	out << "\n";
	for(int i = 1; i <= col; i++){
		out << cDims[i] << " ";
	}
	out << "\n";
	
}
void MixedMatrixP::writeToStreamDelta(ostream &out)
{
	out << "rank of matrix = " << rank << "\n";
	
	if(flag){
		out << "delta(k) : maximum degree of order k subdeterminant " << "\n";
		for(int k = 1; k <= rank; k++){
			out << "delta(" << k << ") = " << delta[k]  << "\n";
		}
	}
	else {
		out << "nu(k) : minimum degree of order k subdeterminant " << "\n";
		for(int k = 1; k <= rank; k++){
			out << "nu(" << k << ") = " << - delta[k]  << "\n";
		}
	}
}

bool MixedMatrixP::checkDimRec(int v, int *chk, int c)
// check the assumption on matrix Q (recursive part)
{
	chk[v] = c;
	if(v < col){
		for(int i = 0; i < row; i++){
			if(!matQ.get(i, v).isZero()){
				int d = matQ.getDim(i, v);
				if(chk[i+col]==0){
					rDims[i+1] = cDims[v+1] + d;
					if(!checkDimRec(i+col, chk, c)){
						return false;
					}
				} else {
					if(rDims[i+1] != (cDims[v+1] + d)) return false;
				}
			}
		}
	} else {
		v = v - col;
		for(int i = 0; i < col; i++){
			if(!matQ.get(v, i).isZero()){
				int d = matQ.getDim(v, i);
				if(chk[i]==0){
					cDims[i+1] = rDims[v+1] - d;
					if(!checkDimRec(i, chk, c)){
						return false;
					}
				} else {
					if(cDims[i+1] != (rDims[v+1] - d)) return false;
				}
			}
		}
	}
	return true;
}

bool MixedMatrixP::checkDim()
// check the assumption on matrix Q (based on dimension analysis)
{
	int *chk = new int[col+row];
	int k = 0;
	for(int i = 0; i < col; i++){
		cDims[i+1] = 0;
		chk[k++] = 0;
	}
	for(int i = 0; i < row; i++){
		rDims[i+1] = 0;
		chk[k++] = 0;
	}
	int c = 0;
	for(int i = 0; i < col; i++){
		if(chk[i]==0){
			c++;
			if(!checkDimRec(i, chk, c)){
				delete [] chk;
				return false;
			}
		}
	}
	for(int i = 0; i < row; i++){
		if(chk[i+col]==0){
			c++;
			if(!checkDimRec(i+col, chk, c)){
				delete [] chk;
				return false;
			}
		}
	}
	
	int *mins = new int[c+1];
	for(int i = 0; i <= c; i++){
		mins[i] = 0;
	}
	for(int i = 0; i < col; i++){
		if(cDims[i+1] < mins[chk[i]]){
			mins[chk[i]] = cDims[i+1];
		}
	}
	for(int i = 0; i < row; i++){
		if(rDims[i+1] < mins[chk[i+col]]){
			mins[chk[i+col]] = rDims[i+1];
		}
	}
	for(int i = 0; i < col; i++){
		cDims[i+1] -= mins[chk[i]];
	}
	for(int i = 0; i < row; i++){
		rDims[i+1] -= mins[chk[i+col]];
	}
	
	delete [] chk;
	return true;
}
void MixedMatrixP::allocMemory()
{
	delta = new int[row+1];
	rDims = new int[row+1];
	cDims = new int[col+1];
}
void MixedMatrixP::allocMemoryG()
{
	phi = new int[row+1];
	psi = new int[col+1];
	
	BTChead = new int[2*BTCnum+1];
	BTCadjlist = new int[BTCnum+1];
	BTCvfirst = new int[vnum+1];
	BTCweight = new int[BTCnum+1];
	BTCpweight = new int[BTCnum+1];
	BPhead = new int[2*BPmax+1];
	BPadjlist = new int[BPmax+1];
	BPvfirst = new int[vnum+1];
	BPweight = new int[BPmax+1];
	vattr = new int[vnum+1];
	
	d = new int[vnum+1];
	p = new int[vnum+1];
	path = new int[vnum+1];
	pathLinfo = new int[vnum+1];
	prev = new int[vnum+1];
	previnfo = new int[vnum+1];
}
void MixedMatrixP::freeMemory()
{
	if(delta!=NULL) delete [] delta;
	delta = NULL;
	if(rDims!=NULL) delete [] rDims;
	rDims = NULL;
	if(cDims!=NULL) delete [] cDims;
	cDims = NULL;
}
void MixedMatrixP::freeMemoryG()
{
	if(phi!=NULL) delete [] phi;
	phi = NULL;
	if(psi!=NULL) delete [] psi;
	psi = NULL;
	
	if(BTChead!=NULL) delete [] BTChead;
	if(BTCadjlist!=NULL) delete [] BTCadjlist;
	if(BTCvfirst!=NULL) delete [] BTCvfirst;
	if(BTCweight!=NULL) delete [] BTCweight;
	if(BTCpweight!=NULL) delete [] BTCpweight;
	if(BPhead!=NULL) delete [] BPhead;
	if(BPadjlist!=NULL) delete [] BPadjlist;
	if(BPvfirst!=NULL) delete [] BPvfirst;
	if(BPweight!=NULL) delete [] BPweight;
	if(vattr!=NULL) delete [] vattr;
	BTChead = NULL;
	BTCadjlist = NULL;
	BTCvfirst = NULL;
	BTCweight = NULL;
	BTCpweight = NULL;
	BPhead = NULL;
	BPadjlist = NULL;
	BPvfirst = NULL;
	BPweight = NULL;
	vattr = NULL;
	
	if(d!=NULL) delete [] d;
	if(p!=NULL) delete [] p;
	if(path!=NULL) delete [] path;
	if(pathLinfo!=NULL) delete [] pathLinfo;
	if(prev!=NULL) delete [] prev;
	if(previnfo!=NULL) delete [] previnfo;
	p = NULL;
	path = NULL;
	pathLinfo = NULL;
	prev = NULL;
	previnfo = NULL;
}
void MixedMatrixP::adjacency_list(int *vfirst, int *head, int *adjlist, int m, int n)
//  making adjacency lists
{
	int v;
	for(v = 1; v <= n; v++){
		vfirst[v] = 0;
	}
	for(int a = m; a > 0; a--){
		int v1 = head[2*a - 1];
		adjlist[a] = vfirst[v1];
		vfirst[v1] = a;
	}
}
MixedMatrixP::~MixedMatrixP()
{
	freeMemory();
	freeMemoryG();
}
void MixedMatrixP::step1(bool b)
{
	flag = b;
	
	vnum = row * 2 + col * 2;
	BTnum = 0;
	for(int i = 0; i < row*col; i++){
		if(!matT.ele[i].isZero()) BTnum++;
	}
	BCnum = row  + col;
	BTCnum = BTnum + BCnum;
	BPmax = row * col;
	freeMemoryG();
	allocMemoryG();
	
	for(int i = 1; i <= row; i++){
		phi[i] = 0;								// initailization of row basis
	}
	for(int i = 1; i <= col; i++){
		psi[i] = 0;								// initailization of column basis
	}
	weight = 0;
	rank = 0;
	int k = 0;
	for(int i = 0; i < row; i++){				// generating BC edges
		k++;
		BTChead[2*k-1] = row + i + 1;
		BTChead[2*k] = i + 1;
		if(flag){
			BTCweight[k] = - rDims[i+1];
		}
		else{
			BTCweight[k] = rDims[i+1];
		}
	}
	for(int i = 0; i < col; i++){				// generating BC edges
		k++;
		BTChead[2*k-1] = row*2 + col + i + 1;
		BTChead[2*k] = row*2 + i + 1;
		if(flag) BTCweight[k] = cDims[i+1];
		else BTCweight[k] = - cDims[i+1];
	}
	for(int i = 0; i < row; i++){				// generating BT edges
		for(int j = 0; j < col; j++){
			if(!matT.get(i,j).isZero()){
				k++;
				BTChead[2*k-1] = row + i + 1;
				BTChead[2*k] = row*2 + j + 1;
				if(flag) BTCweight[k] = - matT.getDim(i,j);
				else BTCweight[k] = matT.getDim(i,j);
			}
		}
	}
	adjacency_list(BTCvfirst, BTChead, BTCadjlist, BTCnum, vnum);
	
	BP1num = 0;
	BPnum = 0;
	for(int i = 1; i <= row; i++){ 				// generating BP1 edges
		if(phi[i]!=0){
			for(int j = 1; j <= col; j++){
				if(psi[j]==0){
					if(!matQ.get(i-1,j-1).isZero()){
						BP1num++;
						BPnum++;
						BPhead[2*BPnum-1] = row*2+col+phi[i];
						BPhead[2*BPnum] = row*2+col+j;
						BPweight[BPnum] = 0;
					}
				}
			}
		}
	}
	BP2num = 0;
	for(int i = 1; i <= row; i++){ 				// generating BP2 edges
		if(phi[i]==0){
			for(int j = 1; j <= col; j++){
				if(psi[j]==0){
					if(!matQ.get(i-1,j-1).isZero()){
						BP2num++;
						BPnum++;
						BPhead[2*BPnum-1] = i;
						BPhead[2*BPnum] = row*2+col+j;
						BPweight[BPnum] = 0;
					}
				}
			}
		}
	}
	BP3num = 0;
	for(int i = 1; i <= row; i++){				// generating BP3 edges
		if(phi[i]!=0){
			BP3num++;
			BPnum++;
			BPhead[2*BPnum-1] = row*2+col+phi[i];
			BPhead[2*BPnum] = i;
			BPweight[BPnum] = 0;
		}
	}
	adjacency_list(BPvfirst, BPhead, BPadjlist, BPnum, vnum);
	
	path_search();
	/*
	if(flag) cout << "initial weight = " << - weight << "\n";
	else cout << "initial weight = " << weight << "\n";
	for(int i = 1; i <= vnum; i++){
		cout << p[i] << " ";
	}
	cout << "\n";
	*/
}
void MixedMatrixP::check_attr()
{
	for(int i = 1; i <= vnum; i++){
		vattr[i] = 0;
	}
	for(int i = 1; i<= row; i++){
		if(BTChead[i*2-1] < BTChead[i*2]){
			vattr[BTChead[i*2]] = 1;
		}
	}
	for(int i = row+1; i<= BCnum; i++){
		if(BTChead[i*2-1] < BTChead[i*2]){
			vattr[BTChead[i*2-1]] = 1;
		}
	}
	for(int i = BCnum+1; i<= BTCnum; i++){
		if(BTChead[i*2-1] > BTChead[i*2]){
			vattr[BTChead[i*2-1]] = 1;
			vattr[BTChead[i*2]] = 1;
		}
	}
}
void MixedMatrixP::step2()
{
	while(true){
		check_attr();
		
		for(int i = 1; i <= BTCnum; i++){
			BTCpweight[i] = BTCweight[i] - p[BTChead[2*i]] + p[BTChead[2*i-1]];
		}
		adjacency_list(BTCvfirst, BTChead, BTCadjlist, BTCnum, vnum);
		
		BP1num = 0;
		BPnum = 0;
		for(int i = 1; i <= row; i++){ // generating BP1 edges
			if(phi[i]!=0){
				for(int j = 1; j <= col; j++){
					if(psi[j]==0){
						if(!matQ.get(i-1,j-1).isZero()){
							BP1num++;
							BPnum++;
							BPhead[2*BPnum-1] = row*2+col+phi[i];
							BPhead[2*BPnum] = row*2+col+j;
							BPweight[BPnum] = - p[BPhead[2*BPnum]] + p[BPhead[2*BPnum-1]];
						}
					}
				}
			}
		}
		BP2num = 0;
		for(int i = 1; i <= row; i++){  // generating BP2 edges
			if(phi[i]==0){
				for(int j = 1; j <= col; j++){
					if(psi[j]==0){
						if(!matQ.get(i-1,j-1).isZero()){
							BP2num++;
							BPnum++;
							BPhead[2*BPnum-1] = i;
							BPhead[2*BPnum] = row*2+col+j;
							BPweight[BPnum] = - p[BPhead[2*BPnum]] + p[BPhead[2*BPnum-1]];
						}
					}
				}
			}
		}
		BP3num = 0;
		for(int i = 1; i <= row; i++){  // generating BP3 edges
			if(phi[i]!=0){
				BP3num++;
				BPnum++;
				BPhead[2*BPnum-1] = row*2+col+phi[i];
				BPhead[2*BPnum] = i;
				BPweight[BPnum] = - p[BPhead[2*BPnum]] + p[BPhead[2*BPnum-1]];
			}
		}
		adjacency_list(BPvfirst, BPhead, BPadjlist, BPnum, vnum);
		
		if(!dijkstra()) break; // there is no directed path to sinks
		
		step3();
		rank++;
		if(flag) delta[rank] = - weight;
		else delta[rank] = weight;
		//cout << "pathLnum = " << pathLnum << "   weight = " << - weight << "\n";
		//cout << "\n";
	}
}
void MixedMatrixP::path_search()
{
	for(int i = 1; i <= vnum; i++){
		if((i <= 2*row) && (i >= row+1)){
			p[i] = 0;
		}
		else p[i] = INFINITY;
	}
	for(int t = 1; t <= vnum; t++){
		for(int i = 1; i <= BTCnum; i++){
			int from = BTChead[i*2-1];
			int to = BTChead[i*2];
			if(p[from] < INFINITY && p[to] > p[from] + BTCweight[i]){
				p[to] = p[from] + BTCweight[i];
			}
		}
		for(int i = 1; i <= BPnum; i++){
			int from = BPhead[i*2-1];
			int to = BPhead[i*2];
			if(p[from] < INFINITY && p[to] > p[from]){
				p[to] = p[from];
			}
		}
	}
}

bool MixedMatrixP::dijkstra()
{
	//cout << "-------------dijkstra--------------\n";
	
	create_heap();
	for(int i = 1; i <= vnum; i++){
		d[i] = INFINITY;
	}
	for(int i = row+1; i <= row*2; i++){
		if(vattr[i]==0){
			d[i] = 0;
			insert_heap(i);
		}
	}
	
	while(true){
		int v = delete_min_heap();
		//cout << v << " ";
		if(v == 0) break;
		
		for(int a = BTCvfirst[v]; a; a = BTCadjlist[a]){
			int w = BTChead[2*a];
			int temp = d[w];
			if(temp > d[v] + BTCpweight[a]){
				d[w] = d[v] + BTCpweight[a];
				if(B[w] == 0) insert_heap(w);
				else decrease_key_heap(w);
				prev[w] = a;
				previnfo[w] = 0;
			}
		}
		for(int a = BPvfirst[v]; a; a = BPadjlist[a]){
			int w = BPhead[2*a];
			int temp = d[w];
			if(temp > d[v] + BPweight[a]){
				d[w] = d[v] + BPweight[a];
				if(B[w] == 0) insert_heap(w);
				else decrease_key_heap(w);
				prev[w] = a;
				previnfo[w] = 1;
			}
		}
	}
	//cout << "\n";
	delete_heap();
	/*
	cout << "d = ";
	for(int i = 1; i <= vnum; i++){
		cout << d[i] << " ";
	}
	cout << "\n";
	cout << "p = ";
	for(int i = 1; i <= vnum; i++){
		cout << p[i] << " ";
	}
	cout << "\n";
	*/
	int min = INFINITY;
	int v = -1;
	for(int i = 2*row+1; i <= 2*row+col; i++){
		if(min > d[i] + p[i] && vattr[i]==0 && d[i]!=INFINITY){
			min = d[i] + p[i];
			v = i;
		}
	}
	if(v == -1) return false;
	
	weight += (d[v] + p[v]);
	pathLnum = 0;
	while(true){
		//cout << v << " ";
		if((vattr[v]==0) && (v <= 2*row) && (v >= row+1)) break;
		else{
			pathLnum++;
			path[pathLnum] = prev[v];
			pathLinfo[pathLnum] = previnfo[v];
			if(previnfo[v] == 0) v = BTChead[prev[v]*2-1];
			else v = BPhead[prev[v]*2-1];
		}
	}
	//cout << "\n";
	
	int dmax = - 1;
	for(int i = 1; i <= vnum; i++){
		if(d[i]!=INFINITY && dmax < d[i]) dmax = d[i];
	}
	for(int i = 1; i <= vnum; i++){
		if(d[i]!=INFINITY) p[i] += d[i];
		else p[i] += dmax;
	}
	
	return true;
}
void MixedMatrixP::step3()
{
	for(int i = pathLnum ; i > 0; i--){
		if(pathLinfo[i] == 0){
			int temp = BTChead[path[i]*2-1];
			BTChead[path[i]*2-1] = BTChead[path[i]*2];
			BTChead[path[i]*2] = temp;
			BTCweight[path[i]] = - BTCweight[path[i]];
		}
		else{
			int r,c;
			if(path[i] <= BP1num){ // BP1
				r = psi[BPhead[path[i]*2-1] - 2*row - col];
				c = BPhead[path[i]*2] - 2*row - col;
				phi[r] = c;
				psi[c] = r;
				psi[BPhead[path[i]*2-1] - 2*row - col] = 0;
				RealNumber one(1);
				RealNumber w = one/matQ.get(r-1,c-1);
				for(int k = 1; k <= row; k++){
					if(k==r) continue;
					RealNumber w2 = w * matQ.get(k-1, c-1);
					for(int l = 1; l <= col; l++){
						RealNumber tmp = matQ.get(k-1, l-1) - w2 * matQ.get(r-1, l-1);
						matQ.put(k-1, l-1, tmp);
					}
				}
			}
			else if(path[i] <= BP1num + BP2num){ // BP2
				r = BPhead[path[i]*2-1];
				c = BPhead[path[i]*2] - 2*row - col;
				phi[r] = c;
				psi[c] = r;
				RealNumber one(1);
				RealNumber w = one/matQ.get(r-1,c-1);
				for(int k = 1; k <= row; k++){
					if(k==r) continue;
					RealNumber w2 = w * matQ.get(k-1, c-1);
					for(int l = 1; l <= col; l++){
						RealNumber tmp = matQ.get(k-1, l-1) - w2 * matQ.get(r-1, l-1);
						matQ.put(k-1, l-1, tmp);
					}
				}
			}
			else{ // BP3
				phi[BPhead[path[i]*2]] = 0;
				psi[BPhead[path[i]*2-1] - 2*row - col] = 0;
			}
		}
	}
	/*
	cout << "\n";
	for(int i = 0; i < row; i++){
		for(int j = 0; j < col; j++){
			cout << setw(5) << matQ.get(i,j);
			cout << "    ";
		}
		cout << "\n";
	}
	for(int i = 1; i<= row; i++){
		cout << phi[i] << " ";
	}
	cout << "\n";
	for(int i = 1; i<= col; i++){
		cout << psi[i] << " ";
	}
	cout << "\n";
	*/
}



void MixedMatrixP::create_heap()
{
	A = new int[vnum+1];
	B = new int[vnum+1];
	for(int i = 1; i <= vnum; i++){
		B[i] = 0;
	}
	size = 1;
}

void MixedMatrixP::insert_heap(int v)
{
	A[size] = v;
	B[v] = size;
	shift_up(size);
	size++;
}

void MixedMatrixP::decrease_key_heap(int v)
{
	shift_up(B[v]);
}

int MixedMatrixP::delete_min_heap()
{
	int min;
	
	if(size > 1){
		min = A[1];
		A[1] = A[size-1];
		B[A[1]] = 1;
		size--;
		shift_down(1);
		return min;
	}
	else return 0;
}

void MixedMatrixP::shift_up(int pp)
{
	int temp;
	int q;
	
	if(pp >= 2){
		q = pp/2;
		if(d[A[pp]] < d[A[q]]){
			temp = A[q];
			A[q] = A[pp];
			A[pp] = temp;
			B[A[pp]] = pp;
			B[A[q]] = q;
			shift_up(q);
		}
	}
}

void MixedMatrixP::shift_down(int pp)
{
	int temp;
	int q;
	
	if(2*pp <= size){
		q = 2*pp;
		if((q+1 <= size) && (d[A[q+1]] < d[A[q]])) q++;
		if(d[A[pp]] > d[A[q]]){
			temp = A[q];
			A[q] = A[pp];
			A[pp] = temp;
			B[A[pp]] = pp;
			B[A[q]] = q;
			shift_down(q);
		}
	}
}

void MixedMatrixP::delete_heap(){
	if(A!=NULL) delete [] A;
	A = NULL;
	if(B!=NULL) delete [] B;
	B = NULL;
	size = 0;
}
