/*
 * MixedMatrixP.hpp	(VIAP)
 * Copyright (C) 2004 Yuji MATSUOKA
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *
 *
 * Yuji MATSUOKA:
 *   mailto:yuji_matsuoka@mist.i.u-tokyo.ac.jp
 *   http://www.sr3.t.u-tokyo.ac.jp/~matsuoka/
 */
#ifndef _MIXEDMATRIXP_HPP_
#define _MIXEDMATRIXP_HPP_

#include <iostream>
#include <iomanip>
using namespace std;
#include <stdlib.h>
#include <time.h>

#include "RealNumber.hpp"			// real number
#include "RealMatrixP.hpp"			// polynomial matrix of real numbers
#include "Variable.hpp"				// independent variables
#include "VariableMatrixP.hpp"		// polynomial matrix of independent variables


class MixedMatrixP {
public:
	int row;							// number of rows
	int col;							// number of columns
	RealMatrixP matQ;					// polynomial matrix of integers
	VariableMatrixP matT;				// polynomial matrix of independent variables
	bool flag;

	int *rDims;						// dimension of rows
	int *cDims;						// dimension of columns
	int rank;						// rank of matrix
	int weight;						// weight of matching
	int *delta;						// delta[k]: optimal weight of rank k matching

	int *phi;						// bases in rows
	int *psi;						// bases in columns

	int vnum;						// number of vertices
	int BTCnum;						// BCnum + BTnum
	int BCnum;						// number of BC edges
	int BTnum;						// number of BT edges
	int BPnum;						// number of BP edges
	int BPmax;						// maximum number of BPnum
	int BP1num;						// number of BP1 edges
	int BP2num;						// number of BP1 edges
	int BP3num;						// number of BP1 edges
	int Mnum;						// number of matching
	int Mmax;						// maximum number of Mnum
	
	int *BTChead;					// information of edges of BP
	int *BTCadjlist;
	int *BTCvfirst;
	int *BTCweight;
	int *BTCpweight;
	int *BPhead;
	int *BPadjlist;
	int *BPvfirst;
	int *BPweight;
	int *vattr;						// vertex attribution
	
	int pathLnum;
	int *path;
	int *pathLinfo;
	int *p;
	int *d;
	int *prev;
	int *previnfo;
	
	MixedMatrixP();
	void init(int r, int c);
	void readFromStream(istream &in);
	void writeToStream(ostream &out);
	void writeToStreamDelta(ostream &out);
	void allocMemory();
	void freeMemory();
	void allocMemoryG();
	void freeMemoryG();
	~MixedMatrixP();
	
	void step1(bool b);
	void step2();
	void check_attr();
	void step3();
	void path_search();
	void adjacency_list(int *vfirst, int *head, int *adjlist, int m, int n);
	bool dijkstra();
	
	bool checkDim();
	bool checkDimRec(int v, int *chk, int c);
	
	int *A;
	int *B;
	int size;
	
	void create_heap();
	void delete_heap();
	void insert_heap(int v);
	void decrease_key_heap(int v);
	int delete_min_heap();
	void shift_up(int pp);
	void shift_down(int pp);
	
};


#endif
