/*
 * SNF.cpp	(SNF)
 * Copyright (C) 2004 Yuji MATSUOKA
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *
 *
 * Yuji MATSUOKA:
 *   mailto:yuji_matsuoka@mist.i.u-tokyo.ac.jp
 *   http://www.sr3.t.u-tokyo.ac.jp/~matsuoka/
 */

int SNF(MixedMatrixP mmp){
	int row = mmp.row;
	int col = mmp.col;
	
	MixedMatrix mm;
	mm.init(row, row, col*2);
	for(int i = 0; i < row; i++){
		for(int j = 0; j < col; j++){
			if(i == j){
				RationalNumber rn(1);
				mm.matQ.put(i, j, rn);
			}
			else{
				RationalNumber rn(0);
				mm.matQ.put(i, j, rn);
			}
		}
		for(int j = 0; j < col; j++){
			int d;
			d = (int) mmp.matQ.get(i,j).doubleValue();
			RationalNumber rn(d);
			mm.matQ.put(i, j + col, rn);
		}
	}
	int n = 1;
	
	for(int i = 0; i < row; i++){
		for(int j = 0; j < col; j++){
			if(i == j){
				Variable var(n);
				mm.matT.put(i, j, var);
				n++;
			}
			else{
				Variable var(0);
				mm.matT.put(i, j, var);
			}
		}
		for(int j = 0; j < col; j++){
			if(!mmp.matT.get(i,j).isZero()){
				Variable var(n);					// n = 0 means zero
				mm.matT.put(i, j + col, var);
				n++;
			}
			else{
				Variable var(0);
				mm.matT.put(i, j + col, var);
			}
		}
	}
	//cout << "input------------------------\n";
	//mm.writeToStream(cout);
	mm.step1();
	mm.step2(true);
	//cout << "CCF------------------------\n";
	//mm.writeToStreamCCF(cout);
	
	cout << "rank = " << mm.rank << "\n";
	cout << "\n";
	
	
	MixedMatrixP MMP;
	MMP.init(row*2, col*2);
	n = 1;
	for(int i = 0; i < row; i++){				// elements of Q (integer)
		for(int j = 0; j < col; j++){
			if(i == j){
				RealNumber rn(1);
				MMP.matQ.put(i, j, rn);
				MMP.matQ.putDim(i, j, 0);
			}
			else{
				RealNumber rn(0);
				MMP.matQ.put(i, j, rn);
			}
			Variable var(0);
			MMP.matT.put(i,j,var);
		}
		for(int j = 0; j < col; j++){
			if(!mmp.matQ.get(i,j).isZero()){
				MMP.matQ.put(i, j + col, mmp.matQ.get(i,j));
				MMP.matQ.putDim(i, j + col, mmp.matQ.getDim(i,j));
			}
			else{
				RealNumber rn(0);
				MMP.matQ.put(i, j + col, rn);
			}
			Variable var(0);
			MMP.matT.put(i,j + col,var);
		}
	}
	for(int i = 0; i < row; i++){				// elements of T
		for(int j = 0; j < col; j++){
			RealNumber rn(0);
			MMP.matQ.put(i + row, j, rn);
			if(i == j){
				Variable var(n);
				n++;
				MMP.matT.put(i + row, j, var);
				MMP.matT.putDim(i, j, 0);
			}
			else{
				Variable var(0);
				MMP.matT.put(i + row, j, var);
				MMP.matT.putDim(i, j, 0);
			}
		}
		for(int j = 0; j < col; j++){
			RealNumber rn(0);
			MMP.matQ.put(i + row, j + col, rn);
			if(!mmp.matT.get(i,j).isZero()){
				Variable var(n);
				n++;
				MMP.matT.put(i + row, j + col, var);
				MMP.matT.putDim(i, j, mmp.matT.getDim(i,j));
			}
			else{
				Variable var(0);
				MMP.matT.put(i + row, j + col, var);
				MMP.matT.putDim(i, j, 0);
			}
		}
	}
	MMP.checkDim();
	/*cout << "r_i = ";
	for(int i = 0; i < row; i++){
		cout << MMP.rDims[i+1] << " ";
	}
	cout << "\n" << "c_j = ";
	for(int j = 0; j < col; j++){
		cout << MMP.cDims[j+1] << " ";
	}
	cout << "\n";
	*/
	//MMP.writeToStream(cout);
	
	
	
	mmp.step1(false);
	mmp.step2();
	//mmp.writeToStreamDelta(cout);
	mmp.delta[0] = 0;
	cout << "e_k(s) : kth monic invariant factor" << "\n";
	for(int k = 1; k < mmp.rank; k++){
		cout << "e_" << k << "(s) = s^" << mmp.delta[k] - mmp.delta[k-1] << "\n";
	}
	
	int *ppr = new int[2*row];
	int *ppc = new int[2*col];
	for(int i = 0; i < row; i++){
		ppr[mm.Pr[i+1] - 1] = i;
		ppr[mm.Pr[i+1+row] - 1] = i+row;
	}
	for(int i = 0; i < col*2; i++){
		ppc[mm.Pc[i+1] - 1] = i;
	}
	/*
	for(int i=0; i < 2*row; i++){
		cout << ppr[i] << " ";
	}
	cout << "\n";
	for(int i=0; i < 2*col; i++){
		cout << ppc[i] << " ";
	}
	cout << "\n";
	*/
	
	
	int maxdeg = 0;
	int mindeg = 0;
	int r = 0;
	int c = 0;
	int i = 0;
	int j = 0;
	
	if(mm.Rs[0]*mm.Cs[0]>0){
		//cout << "calc of D_0\n";
		r = mm.Rs[0];
		c = mm.Cs[0];
		MixedMatrixP mp;
		mp.init(r,c);
		for(i = 0 ; i < r; i++){
			for(j = 0; j < c; j++){
				double vv = mm.matCCFQ.get(i,j).doubleValue();
				mp.matQ.put(i,j,vv);
				mp.matQ.putDim(i,j,MMP.rDims[ppr[i]+1]-MMP.cDims[ppc[j]+1]);
				mp.matT.put(i,j,MMP.matT.get(ppr[i],ppc[j]));
				mp.matT.putDim(i,j,MMP.matT.getDim(ppr[i],ppc[j]));
			}
		}
		for(int k = 0; k < r; k++){
			mp.rDims[k+1] = MMP.rDims[ppr[k]+1];
		}
		for(int k = 0; k < c; k++){
			mp.cDims[k+1] = MMP.cDims[ppc[k]+1];
		}
		//mp.writeToStream(cout);
		mp.step1(false);
		mp.step2();
		maxdeg += mp.delta[r];
		mindeg += mp.delta[r];
		//cout << "r = " << r << " c = " << c << " rank = " << mp.rank <<" nu = " << mp.delta[r] << "\n";
	}
	int b;
	for(b = 1; b <= mm.blocknum; b++){
		//cout << "calc of D_" << b << "\n";
		int s = mm.Rs[b];
		MixedMatrixP mp;
		mp.init(s,s);
		for(i = r; i < r+s; i++){
			for(j = c; j < c+s; j++){
				double vv = mm.matCCFQ.get(i,j).doubleValue();
				mp.matQ.put(i-r,j-c,vv);
				mp.matQ.putDim(i-r,j-c,MMP.rDims[ppr[i]+1]-MMP.cDims[ppc[j]+1]);
				mp.matT.put(i-r,j-c,MMP.matT.get(ppr[i],ppc[j]));
				mp.matT.putDim(i-r,j-c,MMP.matT.getDim(ppr[i],ppc[j]));
			}
		}
		for(int k = 0; k < s; k++){
			mp.rDims[k+1] = MMP.rDims[ppr[r+k]+1];
		}
		for(int k = 0; k < s; k++){
			mp.cDims[k+1] = MMP.cDims[ppc[c+k]+1];
		}
		//mp.writeToStream(cout);
		mp.step1(false);
		mp.step2();
		mindeg += - mp.delta[s];
		//cout << "r = " << s << " c = " << s << " rank =" << mp.rank << " nu = " << mp.delta[s] << "\n"; 
		
		mp.init(s,s);
		for(i = r; i < r+s; i++){
			for(j = c; j < c+s; j++){
				double vv = mm.matCCFQ.get(i,j).doubleValue();
				mp.matQ.put(i-r,j-c,vv);
				mp.matQ.putDim(i-r,j-c,MMP.rDims[ppr[i]+1]-MMP.cDims[ppc[j]+1]);
				mp.matT.put(i-r,j-c,MMP.matT.get(ppr[i],ppc[j]));
				mp.matT.putDim(i-r,j-c,MMP.matT.getDim(ppr[i],ppc[j]));
			}
		}
		for(int k = 0; k < s; k++){
			mp.rDims[k+1] = MMP.rDims[ppr[r+k]+1];
		}
		for(int k = 0; k < s; k++){
			mp.cDims[k+1] = MMP.cDims[ppc[c+k]+1];
		}
		r += s;
		c += s;
		//mp.writeToStream(cout);
		mp.step1(true);
		mp.step2();
		maxdeg += - mp.delta[s];
		//cout << "r = " << s << " c = " << s << " rank =" << mp.rank << " delta = " << mp.delta[s] << "\n"; 
		
	}
	if(mm.Rs[b]*mm.Cs[b]>0){
		//cout << "calc of D_inf\n";
		MixedMatrixP mp;
		mp.init(mm.Rs[b],mm.Cs[b]);
		for(i = r; i < r+mm.Rs[b]; i++){
			for(j = c; j < c+mm.Cs[b]; j++){
				double vv = mm.matCCFQ.get(i,j).doubleValue();
				mp.matQ.put(i-r,j-c,vv);
				mp.matQ.putDim(i-r,j-c,MMP.rDims[ppr[i]+1]-MMP.cDims[ppc[j]+1]);
				mp.matT.put(i-r,j-c,MMP.matT.get(ppr[i],ppc[j]));
				mp.matT.putDim(i-r,j-c,MMP.matT.getDim(ppr[i],ppc[j]));
			}
		}
		for(int k = 0; k < mm.Rs[b]; k++){
			mp.rDims[k+1] = MMP.rDims[ppr[r+k]+1];
		}
		for(int k = 0; k < mm.Cs[b]; k++){
			mp.cDims[k+1] = MMP.cDims[ppc[c+k]+1];
		}
		r += mm.Rs[b];
		c += mm.Cs[b];
		//mp.writeToStream(cout);
		mp.step1(false);
		mp.step2();
		//mp.writeToStreamDelta(cout);
		mindeg += mp.delta[mm.Cs[b]];
		maxdeg += mp.delta[mm.Cs[b]];
		//cout << "r = " << mm.Rs[b] << " c = " << mm.Cs[b] << " rank =" << mp.rank << " nu = " << mp.delta[mm.Cs[b]] << "\n"; 
	}
	
	cout << "highest degree of e_" << mmp.rank << "(s) = " << maxdeg << "\n";
	cout << "lowest degree of e_" << mmp.rank << "(s) = " << mindeg << "\n";
	
	return 0;
}
